from .decorators import ValidationSchema
from .schema import InputField


class DefaultRules:
	date = InputField.Date(regex="^\d{4}\-(0?[1-9]|1[012])\-(0?[1-9]|[12][0-9]|3[01])$")
	pk = InputField.String(required=True)
	gender = InputField.String(allowed=['M', 'F', 'O'])

	@staticmethod
	def pk_validation_rule(pk_name: str):
		return ValidationSchema(**{pk_name: InputField.String(required=True)})

	@staticmethod
	def search_validation_rule(fields: InputField.List = None, filters: InputField.Dictionary = None, **kwargs):
		return ValidationSchema(**{
			"fields": InputField.List(required=True) if not isinstance(fields, InputField.List) else fields,
			"filters": InputField.Dictionary(required=True) if not isinstance(filters, InputField.Dictionary) else filters,
			"limit": InputField.Integer(required=True, nullable=True),
			"offset": InputField.Integer(required=True, nullable=True),
			"order_by": InputField.String(required=False, nullable=True),
			"order_field": InputField.String(required=False, nullable=True),

		})



class ErrorViews:

	@staticmethod
	def __render(request, template_name, context):
		from django.shortcuts import render
		return render(request=request, template_name='errors/error_bad_request.html', context={})

	@staticmethod
	def error_400_handler(request, exception):
		return ErrorViews.__render(request=request, template_name='errors/error_bad_request.html', context={})

	@staticmethod
	def error_403_handler(request, exception):
		return ErrorViews.__render(request=request, template_name='errors/error_permission_denied.html',
		                           context={})

	@staticmethod
	def error_404_handler(request, exception):
		from django.http import HttpResponseNotFound
		content_type = 'text/html'
		return HttpResponseNotFound("no found", content_type=content_type)

	@staticmethod
	def error_500_handler(request):
		return ErrorViews.__render(request=request, template_name='errors/error_internal_server.html',
		                           context={})



from .permissions import (Permissions)
from .users import TokenUser
from .decorators import (require_role, api_permissions)
from .session import SessionData

__all__ = ['Permissions','TokenUser','require_role', 'api_permissions']


# Choices
CHOICE_STATUS = (
	('0', 'Inactive'),
	('1', 'Active')
)

# default column names

COLUMN_NAME_ACTIVE = 'active'
COLUMN_NAME_DELETED = 'is_deleted'

COLUMN_NAME_DELETED_BY = 'deleted_by'
COLUMN_NAME_DELETED_AT = 'deleted_at'

COLUMN_NAME_CREATED_BY = 'created_by'
COLUMN_NAME_CREATED_AT = 'created_at'

COLUMN_NAME_UPDATED_BY = 'updated_by'
COLUMN_NAME_UPDATED_AT = 'updated_at'

DEFAULT_APP_LABLES_TO_ROUTE: set = set()
DEFAULT_DATABASE_NAME: str = 'default'
DEFAULT_MIGRATION_DATABASE: list = [DEFAULT_DATABASE_NAME]

DB_ROUTER_OP_LIST = ('write', 'read', 'relation', 'migration')

DEFAULT_QUERY_OFFSET = 0
DEFAULT_QUERY_LIMIT = 500



import logging

LOG_LEVEL_INFO = logging.INFO
LOG_LEVEL_DEBUG = logging.DEBUG
LOG_LEVEL_WARNING = logging.WARNING
LOG_LEVEL_ERROR = logging.ERROR
LOG_LEVEL_CRITICAL = logging.CRITICAL

CHOICES_LOG_LEVELS = (
	(LOG_LEVEL_INFO, 'INFO'),
	(LOG_LEVEL_DEBUG, 'DEBUG'),
	(LOG_LEVEL_WARNING, 'WARNING'),
	(LOG_LEVEL_ERROR, 'ERROR'),
	(LOG_LEVEL_CRITICAL, 'CRITICAL'),
)

LOG_DB = "logs"
LOG_TABLE = "system_logs"
DEFAULT_CONSOLE_HANDLER_NAME = 'console'
DEFAULT_FILE_HANDLER_NAME = 'file'

DEBUG_REQUIRED_FILTER_CLASS = "django.utils.log.RequireDebugTrue"
DEBUG_REQUIRED_FILTER_NAME = 'require_debug_true'

SIMPLE_LOG_FORMAT = '[%(levelname)s] : %(message)s '
SIMPLE_LOG_FILE_FORMAT = '[%(levelname)s] %(asctime)s : %(message)s '
VERBOSE_LOG_FORMAT = "%(levelname)s %(asctime)s %(module)s %(process:d)s %(thread:d)s %(message)s "

LOG_DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

LOG_FILTERS = {
	DEBUG_REQUIRED_FILTER_NAME: {"()": DEBUG_REQUIRED_FILTER_CLASS}
}

DEFAULT_DBLOG_COLUMN_MAPPINGS = {"level": "levelname", "message": "message", }

SYSTEM_DBLOG_COLUMN_MAPPINGS = {
	**DEFAULT_DBLOG_COLUMN_MAPPINGS, "logger": "name", "msg": "msg", "module": "module", "stack_info": "stack_info",
	"filename": "filename", "func_name": "funcName", "thread_name": "threadName", "process_name": "processName",
	"args": "args", "pathname": "pathname", "exc_info": "exc_info", "exc_text": "exc_text", "lineno": "lineno",
}



import logging
import os
from pathlib import Path

from .constants import NameConst
from .exceptions import MsbConfigException


def __get_config_file_type(config_file):
	return os.path.basename(config_file)


def __get_environment_file(config_path: str, main_file: str, local_file: str):
	if not os.path.isdir(config_path):
		raise MsbConfigException.BaseDirDoesNotExists(config_path, desc=config_path)

	local_config_file = Path(config_path).joinpath(local_file)
	main_config_file = Path(config_path).joinpath(main_file)

	if local_config_file.exists():
		return local_config_file.__str__()
	elif main_config_file.exists():
		return main_config_file.__str__()
	else:
		raise MsbConfigException.ConfigFileDoesNotExist(local_config_file, main_config_file)


def __load_configuration_file(config_file: str):
	if NameConst.DOT_ENV_FILE_NAME in config_file:
		from dotenv import load_dotenv
		return load_dotenv(config_file, verbose=True)


def load_config(env_path: str = '', main_file: str = '', local_file: str = ''):
	try:
		config_file = __get_environment_file(config_path=env_path, main_file=main_file, local_file=local_file)
		load_config = __load_configuration_file(config_file=config_file)

		if load_config:
			os.environ[NameConst.ENV_LOAD_STATUS_KEY_NAME] = '1'
		else:
			raise MsbConfigException.ConfigurationLoadingFailed(config_file)
	except Exception as e:
		logging.error(e)
		exit()



from .client import (ApiRequest, ApiRequestParameter, ApiResponseWrapper)
from .dataclasses import (RequestWrapper, HostUrlsConfig, MsbApiResponse)
from .request import (RequestHeaders, RequestInfo)
from .response import (ApiResponse, RestResponse)
from .wrappers import (IntraServiceRequestFactory)

__all__ = [
	"ApiResponse", "ApiResponseWrapper", "MsbApiResponse", "RestResponse",
	"RequestInfo", "RequestHeaders", "RequestWrapper", "RestResponse",
	"HostUrlsConfig", "IntraServiceRequestFactory",
]



from . import model_fields as model_fields
from .config_model import (ConfigurationModelManager, Configuration)
from .metafields import (MsbModelMetaFields)
from .msb_model import (MsbModel)
from .msb_model_manager import (MsbModelManager, )
from .logging_models import (SystemLogModel, LoggingModelManager, )

__all__ = [
	'Configuration',
	'ConfigurationModelManager',
	'MsbModel',
	'MsbModelManager',
	'model_fields',
	'LoggingModelManager',
	'SystemLogModel',
	'MsbModelMetaFields',
]



from msb.exceptions import AppException


class ValidationExceptions:
	class InvalidValidationSchema(AppException):
		_message = 'Validation schema should be of type ValidationSchema.'

	class InvalidValidationInputDict(AppException):
		_message = 'Validation input should be of type dict.'

	class InvalidValidationInputList(AppException):
		_message = 'Validation input should be of type list.'

	class SchemaValidationFailed(AppException):
		_message = 'Failed to validate the input against the given schema.'

	class ErrorFormattingFailed(AppException):
		_message = 'Failed to format the validation errors.'

	class ValidationFailed(AppException):
		_message = 'Failed to validate the request data.'

	class InvalidPayloadException(AppException):
		_message = 'Invalid payload recieved.'

	class InvalidParamsException(AppException):
		_message = 'Invalid request parameters recieved.'

	class InvalidPayloadFormat(AppException):
		_message = 'Malformed request data recieved.'








__all__ = ["ValidationExceptions"]



from django.conf import settings
from django.core.serializers import serialize
from django.db import models as Models
from msb.cipher import Cipher

from .constants import (COLUMN_NAME_DELETED, COLUMN_NAME_DELETED_BY)
from .msb_model_manager import MsbModelManager
from functools import cache

class MsbModel(Models.Model):
	_private_fields: list = []
	_protected_fields: list = []
	_list_field_names: list = []
	_identifier_field: str = ""

	_encrypt_private_fields: bool = settings.MSB_DB_ENCRYPT_PRIVATE_FIELDS
	_encrypt_protected_fields: bool = settings.MSB_DB_ENCRYPT_PROTECTED_FIELDS
	_pk_is_private_fields: bool = settings.MSB_DB_PK_IS_PRIVATE_FIELD
	_hidden_fields: list = []
	_query_properties: list = []

	class Meta:
		abstract = True

	@property
	def pk_name(self):
		return self._meta.pk.attname

	@property
	def pk_value(self):
		return getattr(self, self.pk_name) if self.pk_name is not None else ""

	@property
	def identifier(self):
		return f"{getattr(self, self.identifier_field_name)}" if hasattr(self, self.identifier_field_name) else ""

	@property
	def rows(self):
		return self.objects

	@classmethod
	def add_query_properties(cls, *properties):
		cls._query_properties.extend(properties)

	objects = MsbModelManager()

	@property
	def serialized(self):
		return serialize('python', [self])

	def delete(self, deleted_by=None, using=None, keep_parents=False):
		if hasattr(self, COLUMN_NAME_DELETED):
			setattr(self, COLUMN_NAME_DELETED, True)

		if hasattr(self, COLUMN_NAME_DELETED_BY):
			setattr(self, COLUMN_NAME_DELETED_BY, deleted_by)
		self.save()
		return True

	def recover(self):
		if hasattr(self, COLUMN_NAME_DELETED):
			setattr(self, COLUMN_NAME_DELETED, False)
		self.save()
		return True

	@property
	def encrypted_fields(self) -> list:
		_secure_fields = self._private_fields if isinstance(self._private_fields, list) else []
		if self._pk_is_private_fields:
			_secure_fields.append(self._meta.pk.attname)
		return _secure_fields

	@property
	def encoded_fields(self) -> list:
		return self._protected_fields if isinstance(self._protected_fields, list) else []

	@property
	def identifier_field_name(self):
		if isinstance(self._identifier_field, str) and len(self._identifier_field) > 0:
			return self._identifier_field
		return ''

	@property
	def related_fields(self):
		fields = []
		for field in self._meta.fields:
			if field.get_internal_type() in ['ForeignKey', 'ManyToManyField']:
				fields.append(field.name)
		return fields

	@property
	def list_fields(self) -> dict:
		return self._get_field_values(
			*[self.pk_name, *self._list_field_names] if isinstance(self._list_field_names, list) else []
		)

	def __formatted_field_value(self, field_name: str, field_value: str, force_encrypt: bool = False):
		if field_name in self.encrypted_fields and (force_encrypt or self._encrypt_private_fields):
			return Cipher.encrypt(field_value)
		elif field_name in self.encoded_fields and (force_encrypt or self._encrypt_protected_fields):
			return Cipher.encode_base64(field_value)
		else:
			return field_value

	def _field_value(self, field_name: str, default=None):
		try:
			if "." in field_name:
				field_name, relation_name, *_ = field_name.split(".")
				local_field_value = getattr(self, field_name, default)
				field_value = local_field_value._get_field_value(relation_name, default=None)
			else:
				field_value = getattr(self, field_name, default)
				if isinstance(field_value, MsbModel):
					field_value = field_value._field_value(field_value.pk_name)

		except Exception as e:
			field_value = default
		return field_value

	@cache
	def __should_fetch_field(self, field_name: str, include_properties: bool) -> bool:
		_should_fetch = not any([
			field_name.startswith('_'),
			field_name in self._hidden_fields,
			(isinstance(field_name, property) and not include_properties),
		])
		return  _should_fetch

	def _get_field_values(self, *field_names, default=None, force_encrypt: bool = False, include_properties: bool = False):
		try:
			return {
				field_name.replace('.', '_'): self.__formatted_field_value(
					field_name, self._field_value(field_name, default), force_encrypt
				) for field_name in field_names
				if self.__should_fetch_field(field_name, include_properties)
			}
		except Exception as e:
			return dict()

	def dict(self, encrypted: bool = False, include_properties: bool = False) -> dict:
		_model_fields = [*self.__dict__.keys(), *getattr(self, '_query_properties', [])]
		return self._get_field_values(*_model_fields, force_encrypt=encrypted, include_properties=include_properties)

	def __str__(self):
		return f"<{self.__class__.__name__} [{self.pk_value}]: {self.identifier}>"

	def __unicode__(self):
		return self.__str__()

	def __repr__(self):
		return self.__str__()



